import tkinter as tk
from collections import namedtuple
from gaspium import Component, error


class SpinBox(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)
        self._box = None
        self._button = None
        self._icon = None
        self._strvar_title = tk.StringVar()
        self._strvar_selection = tk.StringVar()

    def build(self, parent=None, title=None, items=None, prompt="- Select -",
              selection=None, readonly=True, on_submit=None, side=tk.LEFT, anchor="nw",
              padx=5, pady=5, expand=False, fill=None):
        # parent
        parent = self._get_parent(parent)
        # body
        self._body = tk.Frame(parent)
        self._body.pack(side=side, anchor=anchor, padx=padx,
                        pady=pady, expand=expand, fill=fill)
        # set title
        if title is not None:
            self._strvar_title.set(title)
            title_label = tk.Label(self._body, textvariable=self._strvar_title)
            title_label.pack(anchor="nw", pady=(0, 2))
        # on_select handler
        if on_submit:
            info = self._gen_info()
            command = lambda event, on_submit=on_submit, info=info: on_submit(info)
        else:
            command = None
        # create OptionMenu widget
        state = "readonly" if readonly else "normal"
        self._spinbox = tk.Spinbox(self._body, textvariable=self._strvar_selection,
                             values=items, state="readonly")
        self._spinbox.pack(fill=fill, expand=expand)
        self._spinbox.config(relief="flat")
        self._spinbox.bind("<Return>", command)
        self._items = items
        # set default selection
        if selection is None:
            self._strvar_selection.set(prompt)
        else:
            self._strvar_selection.set(selection)

    def read(self):
        str_selection = self._strvar_selection.get()
        index = None
        if not self._items:
            return None
        if str_selection in self._items:
            index = self._items.index(str_selection)
        return (index, str_selection)

    def update(self, title=None):
        # update title
        if title is not None:
            self._strvar_title.set(title)

    def _gen_info(self):
        Info = namedtuple("Info", ("app", "pid", "page", "cid", "component"))
        info = Info(self._page.app, self._page.pid, self._page, self._cid, self)
        return info

    def _get_parent(self, parent=None):
        if not parent:
            if self._page.frame:
                return self._page.frame
            else:
                return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent


import tkinter as tk
from collections import namedtuple
from gaspium import Component, error
import megawidget


class Table(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)
        self._table = None
        self._strvar_entry = tk.StringVar()
        self._strvar_title = tk.StringVar()

    def build(self, parent=None, title=None,
              headers=None, data=None, orient=None, megaconfig=None,
              side=tk.LEFT, anchor="nw", padx=5, pady=5,
              expand=False, fill=None):
        # parent
        parent = self._get_parent(parent)
        # body
        self._body = tk.Frame(parent)
        self._body.pack(side=side, anchor=anchor, padx=padx,
                        pady=pady, expand=expand, fill=fill)
        # label title
        if title is not None:
            label = tk.Label(self._body, textvariable=self._strvar_title)
            label.pack(anchor="nw", pady=(0, 2))
            self._strvar_title.set(title)
       # table
        self._table = megawidget.Table(self._body, titles=headers, data=data,
                                       orient=orient, megaconfig=megaconfig)
        self._table.pack(anchor="w", fill=fill)

    def read(self):
        return None

    def update(self, title=None, text=None):
        if title is not None:
            self._strvar_entry.set(title)

    def _gen_info(self):
        Info = namedtuple("Info", ("app", "pid", "page",
                                   "cid", "component"))
        info = Info(self._page.app, self._page.pid,
                    self._page, self._cid, self)
        return info

    def _get_parent(self, parent=None):
        if not parent:
            if self._page.frame:
                return self._page.frame
            else:
                return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent



"""This Page class is defined inside this module."""
import tkinter as tk
import inspect
from viewable import Viewable
from megawidget import ScrollBox, Toast
from gaspium import error
from collections import namedtuple


def get_info_instance_1(app, page, pid):
    """Internal function !"""
    Info = namedtuple("Info", ("app", "page", "pid"))
    return Info(app, page, pid)


def get_info_instance_2(app, page, pid, data):
    """Internal function !"""
    Info = namedtuple("Info", ("app", "page", "pid", "data"))
    return Info(app, page, pid, data)


class Page:
    """
    This is the definition of the Page class
    """
    def __init__(self, pid=None, name="Page",
                 scrolling="vertical",
                 on_open=None, on_close=None):
        """
        Init.

        [parameters]
        - pid: String, the PID (Page ID) of this page should be unique. If you don't set a PID,
        the method 'new_pid' of the App instance will be automatically called to assign a PID
        to this page.
        - name: string, the name of this page.
        - scrolling: should be one of these values: None, "vertical", "horizontal".
        - on_open: the callback to call when this page is opened. This callback must accept
        the 'info' parameter which is a named tuple that contains useful information.
        - on_close: the callback to call when this page is about to close. This callback must accept
         the 'info' parameter which is a named tuple that contains useful information.
        """
        self._pid = pid
        self._name = name
        self._scrolling = scrolling
        self._on_open = on_open
        self._on_close = on_close
        self._app = None
        self._view = None
        self._frame = None
        self._cids_count = 0
        self._components = {}
        self._cache = []
        self._todo_cache = []
        self._hidden = {}
        self._data = {}
        self._name = name


    @property
    def pid(self):
        """
        Returns the page id
        """
        return self._pid

    @pid.setter
    def pid(self, val):
        """
        Set the page id
        """
        if self._pid and val != self._pid:
            raise error.AlreadyDefinedError
        self._pid = val

    @property
    def name(self):
        """
        Returns the page name
        """
        return self._name

    @property
    def scrolling(self):
        """
        Returns the scrolling value
        """
        return self._scrolling

    @property
    def on_open(self):
        """
        Return the on_open callback
        """
        return self._on_open

    @on_open.setter
    def on_open(self, val):
        """
        Set the on_open callback
        """
        self._on_open = val

    @property
    def on_close(self):
        """
        Return the on_close callback
        """
        return self._on_close

    @on_close.setter
    def on_close(self, val):
        """
        Set the on_close callback
        """
        self._on_close = val

    @property
    def app(self):
        """
        Returns the app reference
        """
        return self._app

    @app.setter
    def app(self, val):
        """
        Set the app reference
        """
        if self._app and val != self._app:
            raise error.AlreadyDefinedError
        self._app = val

    @property
    def body(self):
        """
        Return the body reference
        """
        if self._view:
            return self._view.body.box
        return None

    @property
    def frame(self):
        """
        Return the current Frame container for components
        """
        return self._frame

    @frame.setter
    def frame(self, val):
        """
        Update the current Frame container for components
        """
        self._frame = val

    @property
    def components(self):
        """
        Returns the dictionary of components.
        The dictionary keys are CIDs (Component ID)
        """
        return self._components.copy()

    @property
    def data(self):
        """
        Empty dictionary linked to this page. Store inside it whatever you want.
        """
        return self._data

    @property
    def private_open_method(self):
        """
        Internal method !
        Returns the method reference to open this page.
        """
        return self._open

    @property
    def private_close_method(self):
        """Returns the method reference to close this page.
        You shouldn't use this method !"""
        return self._close

    def toast(self, message, duration=1000, blocking=False):
        """
        Displays a toast that will last for x milliseconds (duration)

        [parameters]
        - message: the message string to show
        - duration: integer, milliseconds before the toast closes itself.
        Set None if you don't want this toast to close itself.
        - blocking: boolean to tell if whether you want the flow to block while
        the toast is opened or not.
        """
        if not self._app:
            raise error.PageStateError("You must add this page to the app first")
        toast = Toast(self._app.root, message=message, duration=duration)
        if blocking:
            toast.wait_window()

    def scroll(self, orient="y", value=1):
        """
        Scrolls the page.
        For orient == "x", set value to:
            - 0: to scroll to left
            - 1: to scroll to right
        For orient == "y", set value to:
            - 0: to scroll to top
            - 1: to scroll to bottom

        [parameters]
        - orient: "x" or "y" to scroll horizontally or vertically
        - value: float value between 0 and 1
        """
        if not self._app:
            raise error.PageStateError("You must add this page to the app first")
        if orient == "x":
            self._view.body.xview_moveto(value)
        else:
            self._view.body.yview_moveto(value)

    def new_cid(self):
        """
        Generate a new valid CID (Component ID).
        The CID string follow this pattern: "cid-x", with x being an integer.

        [return value]
        A new valid CID
        """
        self._cids_count += 1
        return "cid-{}".format(self._cids_count)

    def add(self, component_class, cid=None, **config):
        """
        Add a new component to this page.

        [parameters]
        - component_class: the component class, example: 'gaspium.component.Button'
        - cid: the CID (Component ID) of the component. This CID should be unique.
        If you don't set a CID, the method "new_cid" of this page will be automatically called
        to assign a CID to this component
        - **config: the options to configure this component.

        [return value]
        The CID
        """
        if not cid:
            cid = self.new_cid()
        if cid in self._components:
            raise error.DuplicateComponentError
        if not self._view:
            operation = {"name": "add",
                         "data":
                             {
                                 "component_class": component_class,
                                 "cid": cid,
                                 "config": config
                             }
                         }
            self._todo_cache.append(operation)
            return
        component = component_class(self, cid)
        self._components[cid] = component
        self._check_component_keyword(component.build, config)
        component.build(**config)
        return cid

    def read(self, cid, **kwargs):
        """
        Read the contents of a given component already installed on this page.

        [parameters]
        - cid: the CID (Component ID) of the component
        - **kwargs: optional keyword-arguments

        [return value]
        The value returned by the component
        """
        try:
            component = self._components[cid]
        except KeyError as e:
            raise error.ComponentNotFoundError from None
        else:
            self._check_component_keyword(component.read, kwargs)
            return component.read(**kwargs)

    def update(self, cid, **config):
        """
        Update the contents of a given component already installed on this page.

        [parameters]
        - cid: the CID (Component ID) of the component
        - **config: options to configure this component

        [return value]
        Value returned by the component
        """
        try:
            component = self._components[cid]
        except KeyError as e:
            raise error.ComponentNotFoundError from None
        else:
            self._check_component_keyword(component.update, config)
            return component.update(**config)

    def hide(self, cid):
        """
        Hide a given component already installed on this page.

        [parameters]
        - cid: the CID (Component ID) of the component to hide
        """
        try:
            component = self._components[cid]
        except KeyError as e:
            raise error.ComponentNotFoundError from None
        if not component.body:
            return
        if cid in self._hidden:
            return
        component.body.master.config(width=1, height=1)
        before, after = self._before_after_widget(component.body)
        body = component.body
        pack_info = body.pack_info()
        component.body.pack_forget()
        self._hidden[cid] = {"body": body, "pack_info": pack_info,
                             "before": before, "after": after}

    def unhide(self, cid):
        """
        Unhide a previously hidden component

        [parameters]
        - cid: the CID (Component ID)
        """
        try:
            info = self._hidden[cid]
        except KeyError as e:
            return
        before, after = info["before"], info["after"]
        body = info["body"]
        pack_info = info["pack_info"]
        if before:
            pack_info.update(after=before)
        elif after:
            pack_info.update(before=after)
        body.pack(**pack_info)
        del self._hidden[cid]

    def _open(self):
        """This method shouldn't be called by you"""
        if not self._app:
            msg = "Add the page to the App instance first !"
            raise error.PageStateError(msg)
        if not self._view:
            self._view = View(self)
            self._view.build()
        self._view.body.pack(fill=tk.BOTH, expand=1, padx=0,
                             pady=0)
        # consume to-do
        self._consume_todo()

    def _close(self):
        if not self._view or not self._view.body:
            return
        if self._app.caching:
            if self._on_close:
                info = get_info_instance_1(self._app, self, self._pid)
                self._on_close(info)
            self._view.body.pack_forget()
        else:
            self._view.body.pack_forget()
            self._view.destroy()
            self._view = None
            self._components = dict()

    def _before_after_widget(self, widget):
        before = None
        found = False
        after = None
        for item in widget.master.pack_slaves():
            if item is widget:
                found = True
                continue
            if found:
                after = item
                break
            before = item
        return before, after

    def _consume_todo(self):
        for operation in self._todo_cache:
            name, data = operation["name"], operation["data"]
            if name == "add":
                self.add(data["component_class"], data["cid"],
                         **data["config"])
        if self._app.caching:
            self._todo_cache = []

    def _check_component_keyword(self, operation, kwargs):
        sig = inspect.signature(operation)
        valid_keywords = [item for item, _ in sig.parameters.items()]
        for keyword, _ in kwargs.items():
            if keyword not in valid_keywords:
                msg = "Invalid option '{}'. These are valid options: {}"
                msg = msg.format(keyword, ", ".join(valid_keywords))
                raise error.ComponentInterfaceError(msg)


class View(Viewable):
    def __init__(self, page):
        super().__init__()
        self._page = page
        self._master = page.app.body

    @property
    def master(self):
        return self._master

    def _build(self):
        self._body = ScrollBox(self._master,
                               orient=self._page.scrolling)

    def _on_map(self):
        pass

    def _on_destroy(self):
        if self._page.on_close:
            info = get_info_instance_1(self._page.app, self._page, self._page.pid)
            self._page.on_close(info)



import tkinter as tk
import megawidget
from collections import namedtuple
from gaspium import Component, error


class Choice(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)
        self._choice = None
        self._strvar_title = tk.StringVar()

    def build(self, parent=None, title=None, items=None,
              selection=None, flavor="radio",
              stacking="vertical", on_change=None,
              side=tk.LEFT, anchor="nw",
              padx=5, pady=5, expand=False, fill=None):
        # parent
        parent = self._get_parent(parent)
        # body
        self._body = tk.Frame(parent)
        self._body.pack(side=side, anchor=anchor, padx=padx,
                        pady=pady, expand=expand, fill=fill)
        # set title
        if title is not None:
            self._strvar_title.set(title)
            title_label = tk.Label(self._body, textvariable=self._strvar_title)
            title_label.pack(anchor="nw", pady=(0, 2))
        # set on_change
        if on_change:
            info = self._gen_info()
            command = (lambda choice, on_change=on_change, info=info: on_change(info))
        else:
            command = None
        # create choice megawidget
        self._choice = megawidget.Choice(self._body, items=items, selection=selection,
                                         flavor=flavor, stacking=stacking,
                                         on_change=command)
        self._choice.pack(fill=fill, expand=expand)

    def read(self):
        if self._choice:
            return self._choice.selection
        return None

    def update(self, title=None):
        if title is not None:
            self._strvar_title.set(title)

    def _gen_info(self):
        Info = namedtuple("Info", ("app", "pid", "page", "cid", "component"))
        info = Info(self._page.app, self._page.pid, self._page, self._cid, self)
        return info

    def _get_parent(self, parent=None):
        if not parent:
            if self._page.frame:
                return self._page.frame
            else:
                return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent



import tkinter as tk
from collections import namedtuple
from gaspium import Component, error


class Image(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)
        self._strvar_caption = tk.StringVar()
        self._img_id = None
        self._photo_image = None

    def build(self, parent=None, image=None, caption=None,
              width=None, height=None, on_click=None,
              side=tk.LEFT, anchor="nw", padx=5, pady=5,
              expand=False, fill=None):
        # parent
        parent = self._get_parent(parent)
        # body
        self._body = tk.Frame(parent)
        self._body.pack(side=side, anchor=anchor, padx=padx,
                        pady=pady, expand=expand, fill=fill)
        # photo image
        width = 0 if width is None else width
        height = 0 if height is None else height
        self._photo_image = tk.PhotoImage(data=image,
                                    width=width,
                                    height=height)
        self._canvas = tk.Canvas(self._body,
                           width=self._photo_image.width(),
                           height=self._photo_image.height())
        self._canvas.pack()
        self._img_id = self._canvas.create_image(0, 0,
                                                 image=self._photo_image, anchor="nw")
        # label caption
        if caption is not None:
            label = tk.Label(self._body, textvariable=self._strvar_caption)
            label.pack(pady=(2, 0))
            self._strvar_caption.set(caption)
        # on click
        if on_click:
            info = self._gen_info()
            command = (lambda event, on_submit=on_click, info=info: on_submit(info))
            self._canvas.config(cursor="hand1")
        else:
            command = None
        self._canvas.bind("<Button-1>", command)

    def read(self):
        return None

    def update(self, image=None, width=None, height=None, caption=None):
        if image is not None:
            if self._img_id:
                self._canvas.delete(self._img_id)
            width = 0 if width is None else width
            height = 0 if height is None else height
            self._photo_image = tk.PhotoImage(data=image,
                                        width=width,
                                        height=height)
            self._canvas.config(width=self._photo_image.width(),
                                height=self._photo_image.height())
            self._img_id = self._canvas.create_image(0, 0,
                                                     image=self._photo_image, anchor="nw")
        if caption is not None:
            self._strvar_caption.set(caption)

    def _gen_info(self):
        Info = namedtuple("Info", ("app", "pid", "page", "cid", "component"))
        info = Info(self._page.app, self._page.pid, self._page, self._cid, self)
        return info

    def _get_parent(self, parent=None):
        if not parent:
            if self._page.frame:
                return self._page.frame
            else:
                return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent



import tkinter as tk
from collections import namedtuple
from gaspium import Component, error


class Label(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)
        self._box = None
        self._icon = None
        self._label = None
        self._strvar_text = tk.StringVar()

    def build(self, parent=None, text="This is a Label", color="tomato",
              font_family="Liberation Mono", font_size=15, font_weight="bold",
              icon=None, compound=tk.LEFT, on_click=None, side=tk.LEFT,
              anchor="nw", padx=5, pady=5, expand=False, fill=None):
        # parent
        parent = self._get_parent(parent)
        # keep a reference to icon
        self._icon = icon
        # fill text stringvar
        if text:
            self._strvar_text.set(text)
        # font
        font = (font_family, font_size, font_weight)
        # create label
        self._label = tk.Label(parent, textvariable=self._strvar_text,
                         foreground=color, font=font,
                         image=icon, compound=compound)
        self._label.pack(side=side, anchor=anchor, padx=padx, pady=pady,
                          expand=expand, fill=fill)
        self._body = self._label

    def read(self):
        return None

    def update(self, text=None):
        # update text
        if text is not None:
            self._strvar_text.set(text)

    def _gen_info(self):
        Info = namedtuple("Info", ("app", "pid", "page", "cid", "component"))
        info = Info(self._page.app, self._page.pid, self._page, self._cid, self)
        return info

    def _get_parent(self, parent=None):
        if not parent:
            if self._page.frame:
                return self._page.frame
            else:
                return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent



"""Demo. Note: this demo needs an image that is only on my computer."""
from gaspium import App, Page
from gaspium.component import Frame, Label, Choice, Button, Editor, Entry, Table, Image, SpinBox, PathField, OptionMenu
from cyberpunk_theme.widget.button import get_button_dark_style, get_button_dark_filled_style, get_button_blue_style, get_button_blue_filled_style, get_button_green_style, get_button_green_filled_style, get_button_yellow_style, get_button_yellow_filled_style, get_button_red_style, get_button_red_filled_style


TEXT = """\
Hello Friend !

This GUI is built with Gaspium.

Gaspium: high-productivity framework to build Python apps.

This project is part of the Pyrustic Open Ecosystem.



                      	Join the Pyrustic Open Ecosystem !

	                   https://pyrustic.github.io       

"""


def get_home_page(app):
    page = Page(name="Home", scrolling=None)
    # Row 1
    page.add(Frame, cid="frame1")
    # radiobbutons
    items = ("TkF", "Gaspium", "Shared", "TkStyle", "Litemark",
             "Backstage", "Cyberpunk-Theme", "Threadom",
             "Megawidget", "Viewable", "...")
    page.add(Choice, parent="frame1",
             title="Pyrustic Ecosystem", flavor="radio", items=items, selection=1)
    # Editor
    page.add(Editor, parent="frame1", fill="both", expand=True, text=TEXT)
    # Choices
    items = ("Python", "Desktop", "Frontend", "Backend", "Distribution",
             "Megawidgets", "Multithreading   ", "Theme|Style",
             "Autocomplete", "Productivity", "Bullshit", "More...")
    page.add(Choice, parent="frame1", flavor="check", items=items,
             selection=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11))

    # Row 2
    page.add(Frame, cid="frame2")
    # table
    headers = ("Mountain", "Height", "Planet")
    data = [["Everest", "8,848", "Earth"],
            ["Skil Brum", "7,410", "Mars"],
            ["Abi Gamin", "7,355", "Venus"],
            ["Mana Peak", "7,272", "Krypton"],
            ["Karjiang", "7,221", "Tao-Hung"]]
    megaconfig = {"listboxes_columns": {"width": 14, "height": 6}}
    page.add(Table, parent="frame2", headers=headers, data=data,
             megaconfig=megaconfig)
    # image
    page.add(Frame)
    try:
        with open("/home/alex/demo-img.png", "rb") as file:
            img = file.read()
        page.add(Image, parent="frame2", image=img, anchor="center", fill="x", expand=1)
    except Exception as e:
        pass
    # Row 3
    page.add(Frame, cid="frame3")
    # entry
    # --
    page.add(Frame, parent="frame3", side="left")
    page.add(Entry, title="Name", side="top", text="Alex Rustic")
    page.add(Entry, title="Password", secretive=True, text="Password", side="top")
    # --
    page.add(Frame, parent="frame3", side="left")
    page.add(Entry, title="GitHub", side="top", text="github.com/pyrustic", readonly=True)
    page.add(OptionMenu, title="Options", side="top", items=("Item 1", "Item 2"))
    # --
    page.add(Frame, parent="frame3", side="left")
    page.add(PathField, title="Path", side="top", path="/home/alex/demo")
    page.add(SpinBox, title="Duration", side="top", prompt="3.141592653")
    # --
    page.add(Frame, parent="frame3", side="right", anchor="s")
    page.add(Label, side="right", text="{GASP}", color="#242a2f", anchor="center", font_size=58)

    # Row 4
    page.add(Frame, cid="frame4", side="bottom")
    page.add(Button, parent="frame4", style=get_button_dark_style())
    page.add(Button, parent="frame4", style=get_button_dark_filled_style())
    page.add(Button, parent="frame4", style=get_button_blue_style())
    page.add(Button, parent="frame4", style=get_button_blue_filled_style())
    page.add(Button, parent="frame4", style=get_button_green_style())
    page.add(Button, parent="frame4", style=get_button_green_filled_style())
    page.add(Button, parent="frame4", style=get_button_yellow_style())
    page.add(Button, parent="frame4", style=get_button_yellow_filled_style())
    page.add(Button, parent="frame4", style=get_button_red_style())
    page.add(Button, parent="frame4", style=get_button_red_filled_style())
    return page


def get_process_page(app):
    page = Page(name="Process")
    return page


def get_about_page(app):
    page = Page(name="About")
    return page


def get_row_1(app):
    pass


def main():
    """Demo"""
    app = App(width="920", height="537")

    home_page = get_home_page(app)
    app.add(home_page)

    process_page = get_process_page(app)
    app.add(process_page)

    about_page = get_about_page(app)
    app.add(about_page)


    app.start()


if __name__ == "__main__":
    main()



import tkinter as tk
from collections import namedtuple
from gaspium import Component, error


class Button(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)
        self._button = None
        self._icon = None
        self._strvar_text = tk.StringVar()

    def build(self, parent=None, text="Submit", icon=None,
              compound=tk.LEFT, on_click=None, style=None,
              state="normal", side=tk.LEFT, anchor="nw",
              padx=5, pady=5, expand=False, fill=None):
        # parent
        parent = self._get_parent(parent)
        # keep a reference to icon
        self._icon = icon
        # fill text stringvar
        if text:
            self._strvar_text.set(text)
        # on_click handler
        if on_click:
            info = self._gen_info()
            command = lambda on_click=on_click, info=info: on_click(info)
        else:
            command = None
        # create button
        self._button = tk.Button(parent, textvariable=self._strvar_text,
                                 command=command, image=icon, compound=compound,
                                 cursor="hand1", state=state)
        self._button.pack(side=side, anchor=anchor, padx=padx, pady=pady,
                          expand=expand, fill=fill)
        self._body = self._button
        # styling
        if style:
            style.target(self._button)

    def read(self):
        return None

    def update(self, text=None, style=None,
               state=None):
        # update text
        if text is not None:
            self._strvar_text.set(text)
        # update style
        if style is not None:
            style.target(self._button)
        # update state
        if state is not None:
            self._button.config(state=state)

    def _gen_info(self):
        Info = namedtuple("Info", ("app", "pid", "page", "cid", "component"))
        info = Info(self._page.app, self._page.pid, self._page, self._cid, self)
        return info

    def _get_parent(self, parent=None):
        if not parent:
            if self._page.frame:
                return self._page.frame
            else:
                return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent



import tkinter as tk
from tkinter.scrolledtext import ScrolledText
from collections import namedtuple
from gaspium import Component, error
import litemark


class Litemark(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)
        self._box = None
        self._strvar_title = tk.StringVar()
        self._text_widget = None
        self._viewer = None

    def build(self, parent=None, title=None, text=None, path=None, root=None,
              style=None,
              width=45, height=10, side=tk.LEFT, anchor="nw",
              padx=5, pady=5, expand=False, fill=None):
        # parent
        parent = self._get_parent(parent)
        # body
        self._body = tk.Frame(parent)
        self._body.pack(side=side, anchor=anchor, padx=padx,
                        pady=pady, expand=expand, fill=fill)
        # set title
        if title is not None:
            self._strvar_title.set(title)
            title_label = tk.Label(self._body, textvariable=self._strvar_title)
            title_label.pack(anchor="nw", pady=(0, 2))
        # create text widget
        self._text_widget = ScrolledText(self._body, width=width, height=height,
                                         bg="white", wrap="word")
        self._text_widget.pack(fill=fill, expand=expand)
        self._viewer = litemark.Viewer(widget=self._text_widget, root=root,
                                       style=style)
        if text:
            self._viewer.render(text)
        elif path:
            self._viewer.open(path)
        self._viewer.readonly = True

    def read(self):
        if self._text_widget:
            return self._text_widget.get("1.0", tk.END)
        return None

    def update(self, title=None, text=None):
        if title is not None:
            self._strvar_title.set(title)
        if text is not None:
            self._viewer.clear()
            self._viewer.render(text)
            self._viewer.readonly = True

    def _gen_info(self):
        Info = namedtuple("Info", ("app", "pid", "page", "cid", "component"))
        info = Info(self._page.app, self._page.pid, self._page, self._cid, self)
        return info

    def _get_parent(self, parent=None):
        if not parent:
            if self._page.frame:
                return self._page.frame
            else:
                return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent



"""The App class is defined inside this module."""
import sys
import tkinter as tk
from viewable import Viewable
from tkf import App as TkfApp
from cyberpunk_theme import Cyberpunk
from gaspium import error
from gaspium.page import get_info_instance_2


class App:
    """This class is the entry point of your Gaspium app"""
    def __init__(self, title="Application", width=800, height=500,
                 theme=Cyberpunk(), caching=False,
                 resizable=(False, True), on_exit=None):
        """
        Init.

        [parameters]
        - title: string, the title of the app
        - width: int, the width of the app
        - height: int, the height of the app
        - theme: the theme, i.e. an instance of tkstyle.Theme
        - caching: boolean to tell if whether you want pages to be cached or built at open.
        - resizable: tuple of boolean to tell if you want the width and the height of the app to be resizable.
        - on_exit: the on_exit handler, i.e. a function that will be called on exit.
        """
        self._title = title
        self._width = width
        self._height = height
        self._theme = theme
        self._caching = caching
        self._resizable = resizable
        self._on_exit = on_exit
        self._pages = {}
        self._page = None
        self._todo_open_cache = []
        self._todo_menu_cache = []
        self._menubar = None
        self._view = None
        self._menu_map = {}
        self._pids = []
        self._pids_count = 0
        self._tkf_app = TkfApp()
        self._root = self._tkf_app.root
        self._opening = False
        self._busy_closing_page = False
        self._opened_pages_count = 0
        self._started = False
        self._home = None
        self._data = {}
        self._setup()

    @property
    def title(self):
        """
        Return the title of the app
        """
        return self._title

    @title.setter
    def title(self, val):
        """
        Set the title of the app
        """
        if self._started and self._title:
            raise error.AlreadyDefinedError
        self._title = val

    @property
    def width(self):
        """
        Return the width of the app
        """
        return self._width

    @width.setter
    def width(self, val):
        """
        Set the width of the app
        """
        if self._started and self._width:
            raise error.AlreadyDefinedError
        self._width = val

    @property
    def height(self):
        """
        Return the height of the app
        """
        return self._height

    @height.setter
    def height(self, val):
        """
        Set the height of the app
        """
        if self._started and self._height:
            raise error.AlreadyDefinedError
        self._height = val

    @property
    def theme(self):
        """
        Return the current theme
        """
        return self._theme

    @theme.setter
    def theme(self, val):
        """
        Set a theme, i.e., a tkstyle.Theme instance
        """
        if self._started and self._theme:
            raise error.AlreadyDefinedError
        self._theme = val

    @property
    def caching(self):
        """
        Return a boolean to indicate if the caching option is True or False.
        """
        return self._caching

    @caching.setter
    def caching(self, val):
        """
        Set True if you want pages to be cached. Cached pages retains their data.
        """
        if self._started and self._caching:
            raise error.AlreadyDefinedError
        self._caching = val

    @property
    def resizable(self):
        """
        Return the resizable booleans tuple
        """
        return self._tkf_app.resizable

    @resizable.setter
    def resizable(self, val):
        """
        Set the resizable booleans tuple
        """
        if self._started:
            raise error.AlreadyDefinedError
        self._tkf_app.resizable = val

    @property
    def on_exit(self):
        """
        Return the on_exit callback
        """
        return self._on_exit

    @on_exit.setter
    def on_exit(self, val):
        """
        Set the on_exit handler. The handler is a function that accepts no argument
        """
        if self._started and self._on_exit:
            raise error.AlreadyDefinedError
        self._on_exit = val

    @property
    def home(self):
        """
        Return the home page instance, i.e. the first page added to this app
        """
        return self._home

    @property
    def page(self):
        """
        Return the currently opened page instance
        """
        return self._page

    @property
    def pages(self):
        """
        Return the copy of an internal dictionary that contains pages.
        Note: Keys are PIDs (Page ID)
        """
        return self._pages.copy()

    @property
    def root(self):
        """
        Return the root, i.e. the Tkinter's Tk instance that serves as the root of this app.
        """
        return self._root

    @property
    def body(self):
        """
        Return the Body frame of this app, i.e. the tkinter.Frame instance that serves of the body of this app
        """
        if self._view:
            return self._view.body
        return None

    @property
    def data(self):
        """
        This is an empty dictionary linked to this app.
        Use it to store whatever you want.
        Example: you can use it to retain the information about whether the user is authenticated or not."""
        return self._data

    def add(self, page, indexable=True, category=None):
        """
        Add a page instance to the app.

        [parameters]
        - page: an instance of gaspium.page.Page
        - indexable: boolean, if False, the page won't be indexed in the menubar
        - category: string, the menu category name under which the page is indexed

        [return value]
        Returns the pid

        Raises 'gaspium.error.DuplicatePageError' if the pid already exists
        """
        if not page.app:
            page.app = self
        if not page.pid:
            page.pid = self.new_pid()
        pid = page.pid
        if pid in self._pages:
            raise error.DuplicatePageError
        self._pages[pid] = page
        self._pids.append(pid)
        if indexable:
            if self._view.body:
                self._view.populate_menubar(pid, page, category)
            else:
                data = (pid, page, category)
                self._todo_menu_cache.append(data)
        # add the first page as home
        if not self._home:
            self._home = pid
        return pid

    def new_pid(self):
        """
        Generate a new valid PID (Page ID).
        A PID value follows this pattern: 'pid-x', with x being an integer

        [return value]
        A new valid PID
        """
        self._pids_count += 1
        return "pid-{}".format(self._pids_count)

    def open(self, pid, data=None):
        """
        Open a page specified by its PID (Page ID).

        [parameters]
        - pid: PID
        - data: Associated data. This data is passed to the 'on_open' callback of the page.
         Note: If you open this page from the command line, the data passed is split into a tuple.

        [return value]
        Raise gaspium.error.PageNotFoundError if not page is associated to this PID.
        Raise gaspium.error.PageStateError if you try to open a new page inside 'on_close' callback.
        """
        if self._busy_closing_page:
            msg = "Don't open a new page inside 'on_close' callback"
            raise error.PageStateError(msg)
        if not self._view.body:
            self._todo_open_cache.append((pid, data))
            return
        if pid not in self._pages:
            raise error.PageNotFoundError
        self._opening = True
        self._opened_pages_count += 1
        cache = self._opened_pages_count
        if self._page:
            self._busy_closing_page = True
            self._page.private_close_method()
            self._busy_closing_page = False
        self._page = self._pages[pid]
        # on open
        if self._page.on_open:
            info = get_info_instance_2(self, self._page, pid, data)
            self._page.on_open(info)
        if cache == self._opened_pages_count:
            self._page.private_open_method()
        self._opening = False

    def start(self):
        """
        Start the app. Mainloop here.
        """
        self._started = True
        # open home
        if self._home:
            self.open(self._home)
        if self._handle_open_page_from_cli():
            self._tkf_app.start()

    def exit(self):
        """
        Exit the app.
        """
        self._tkf_app.exit()

    def _setup(self):
        # set theme
        if self._theme:
            self._tkf_app.theme = self._theme
        # set width and height
        if self._width and self._height:
            cache = "{}x{}+0+0".format(self._width, self._height)
            self._root.geometry(cache)
        # set resizable
        self.resizable = self._resizable
        # center the app
        self._tkf_app.center()
        # set the main view
        self._view = View(self, self._tkf_app.root,
                          self._todo, self._on_exit)
        self._tkf_app.view = self._view

    def _todo(self):
        for pid, page, category in self._todo_menu_cache:
            self._view.populate_menubar(pid, page, category)
        try:
            pid, data = self._todo_open_cache[-1]
        except IndexError as e:
            pass
        else:
            self.open(pid, data)
        self._todo_menu_cache = []
        self._todo_open_cache = []

    def _handle_open_page_from_cli(self):
        cache = sys.argv[1:]
        if not cache:
            return True
        pid = cache[0]
        data = None
        if len(cache) > 1:
            data = tuple(cache[1:])
        if pid not in self._pages:
            msg = "\nThe page id '{}' isn't defined.\n\nAvailable pages\n===============\n{}\n"
            msg = msg.format(pid, "  ".join(self._pages.keys()))
            print(msg)
            self.exit()
            return False
        self.open(pid, data=data)
        return True


class View(Viewable):

    def __init__(self, app, root, todo_on_map, on_exit):
        super().__init__()
        self._app = app
        self._todo_on_map = todo_on_map
        self._on_exit = on_exit
        self._on_exit = on_exit
        self._root = root
        self._body = None
        self._menubar = None
        self._menu_cache = []
        self._menu_map = {}

    def _build(self):
        self._body = tk.Frame(self._root)
        # set menubar
        self._menubar = tk.Menu(self._root)
        self._root.config(menu=self._menubar)

    def _on_map(self):
        self._todo_on_map()
        self._set_title()

    def populate_menubar(self, pid, page, category):
        page_name = page.name
        command = (lambda self=self, pid=pid:
                   self._app.open(pid))
        if category:
            if category in self._menu_map:
                menu = self._menu_map[category]
            else:
                menu = tk.Menu(self._menubar, tearoff=0)
                self._menu_map[category] = menu
                self._menubar.add_cascade(label=category,
                                          menu=menu)
            menu.add_command(label=page_name,
                             command=command)
        else:
            if page_name in self._menu_map:
                menu = self._menu_map[page_name]
            else:
                menu = self._menubar
            menu.add_command(label=page_name,
                             command=command)

    def _on_destroy(self):
        if self._on_exit:
            self._on_exit(self)

    def _set_title(self):
        title = "{} | Pyrustic Gaspium".format(self._app.title)
        self._root.title(title)



class Error(Exception):
    pass


class PageNotFoundError(Error):
    pass


class DuplicatePageError(Error):
    pass


class PageStateError(Error):
    pass


class AlreadyDefinedError(Error):
    pass


class DuplicateComponentError(Error):
    pass


class ComponentNotFoundError(Error):
    pass


class ComponentInterfaceError(Error):
    pass



"""This module exposes the default components classes (Button, Frame, Table, et cetera.)"""
from gaspium.component.frame import Frame
from gaspium.component.button import Button
from gaspium.component.choice import Choice
from gaspium.component.entry import Entry
from gaspium.component.optionmenu import OptionMenu
from gaspium.component.editor import Editor
from gaspium.component.image import Image
from gaspium.component.label import Label
from gaspium.component.litemark import Litemark
from gaspium.component.pathfield import PathField
from gaspium.component.spinbox import SpinBox
from gaspium.component.table import Table
#  TODO: add Slider and Tree

__all__ = ["Frame", "Button", "Choice", "Entry", "OptionMenu",
           "Editor", "Image", "Label", "Litemark", "PathField",
           "SpinBox", "Table"]



"""Main module exposing the classes 'App', 'Page', and the module 'component' that contains default components."""
from gaspium.app import App
from gaspium.page import Page
from gaspium.page import error


__all__ = ["App", "Page", "Component"]


class Component:
    """
    Base class to create a Component class like the ones in 'gaspium.component'.
    """
    def __init__(self, page, cid):
        self._page = page
        self._cid = cid
        self._body = None
        self._parts = dict()

    @property
    def page(self):
        """
        Get the page instance.
        """
        return self._page

    @property
    def cid(self):
        """
        Get the CID (Component ID).
        """
        return self._cid

    @property
    def body(self):
        """
        Get the body of this component.
        """
        return self._body

    @property
    def parts(self):
        """
        Get the parts of this component.
        """
        return self._parts.copy()

    def build(self, **config):
        """
        Method to call to build this component.
        This method must create the body of this component and assign it to 'self._body'.
        This method should also update the 'parts' property.

        [parameters]
        - **config: options to configure this component.
         This configuration include the five keywords of the layout mechanism
         ('parent', 'side', 'anchor', 'fill', and 'expand').
        """
        pass

    def read(self, **kwargs):
        """
        Method to read the contents of this component.

        [parameters]
        - **kwargs: keyword-arguments.

        [return value]
        The contents of this component.
        """
        pass

    def update(self, **config):
        """
        Method to update this component.

        [parameters]
        - **config: options to configure this component.

        [return value]
        None or something else.
        """
        pass



import tkinter as tk
from gaspium import Component, error


class Frame(Component):
    def __init__(self, page, cid):
        super().__init__(page, cid)

    def build(self, color=None, width=None, height=None,
              parent=None, side=tk.TOP, anchor=tk.CENTER,
              padx=0, pady=0, expand=False, fill=tk.X):
        parent = self._get_parent(parent)
        self._body = tk.Frame(parent, background=color, width=width,
                              height=height)
        self._body.pack(side=side, anchor=anchor, padx=padx, pady=pady,
                        expand=expand, fill=fill)
        self._page.frame = self._body

    def read(self):
        pass

    def update(self, color=None):
        if color is not None:
            self._body.config(background=color)

    def _get_parent(self, parent=None):
        if not parent:
            return self._page.body
        if isinstance(parent, str):
            component = self._page.components.get(parent)
            if not component:
                msg = "Component ID '{}' not found !".format(parent)
                raise error.ComponentNotFoundError(msg)
            return component.body
        return parent



"""This module runs the tests located in the package tests"""
import os
from backstage import api


project_dir = os.getcwd()
tests_success, tests_result = api.run_tests(project_dir)
if tests_success is True:
    print("Testing passed !\n")
elif tests_success is False:
    print("Testing failed !\n")
    print(tests_result)
else:
    print("No tests available.")
